/*	
 *	clipapu.h
 *
 *	P/ECE APU (RICOH RP2A03) Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2005 Naoyuki Sawa
 *
 *	* Sun Feb 06 18:39:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#ifndef __CLIP_APU_H__
#define __CLIP_APU_H__

/****************************************************************************
 *
 ****************************************************************************/

/* APU_ASMV{`ǂݍ݂܂B
 * APU_ASMV{̖`/`ɂC/AZuR[h̐؂ւ́A
 * C\[Xt@C(*.c)ƃAZu\[Xt@C(*.s)QƂ̂ŁA
 * ʂ̃CN[ht@CɕĂ܂B
 * (CL̒`܂ރCN[ht@Ĉ܂܃AZu\[X
 *  CN[hƁAAZuG[ɂȂĂ܂߁B)
 */
#include "clipapua.h"

/********** ȍ~͏Ȃł **********/

/* P/ECEȊOł̓AZuR[h͎gpł܂B
 */
#ifndef PIECE
#undef APU_ASM
#endif /*PIECE*/

/****************************************************************************
 *
 ****************************************************************************/

#define APU_CLOCK (NTSC_COLOR_SUBCARRIER / 2)
#define APUBUFLEN 320 /* = 16000/50 c 1[tick](=1/50[sec])̃Tv */

/* OXJE^ (60Hz) */
typedef struct _APULENGTH {
	unsigned char flags;		/* D0: 0 = OXJE^ */
					/*     1 = OXJE^L */
	unsigned char count;		/* OXJE^ (0..127) */
} APULENGTH;

/* jAJE^ (240Hz) */
typedef struct _APULINEAR {
	unsigned char flags;		/* D0: 0 = jAJE^ */
					/*     1 = jAJE^L */
	unsigned char count;		/* jAJE^ (0..127) */
} APULINEAR;

/* Gx[v (240Hz) */
typedef struct _APUENVELOPE {
	unsigned char flags;		/* D0: 0 = Gx[v */
					/*     1 = Gx[vL */
					/* D1: 0 = [v */
					/*     1 = [vL */
	unsigned char period;		/* XV (1..16) */
	unsigned char count;		/* isJE^ (0..16) */
	unsigned char volume;		/* {[ (0<<3..15<<3) */
} APUENVELOPE;

/* XC[v (120Hz) */
typedef struct _APUSWEEP {
	unsigned char flags;		/* D0: 0 = XC[v */
					/*     1 = XC[vL */
					/* D1: 0 =  */
					/*     1 =  */
	unsigned char shift;		/* gʂ̃Vtg */
	unsigned char period;		/* XV (1..8) */
	unsigned char count;		/* isJE^ (0..8) */
	short wavelength;		/* g ($000..$800) */
} APUSWEEP;

/* `g (Ch0,Ch1) */
typedef struct _APUSQUARE {
	APULENGTH length;		/* OXJE^ */
	APUENVELOPE envelope;		/* Gx[v */
	APUSWEEP sweep;			/* XC[v */
	unsigned char duty;		/* f[eBTCN (2,4,8,12) */
	unsigned char waveindex;	/* g`CfNX (0..15) */
	int wavecount;			/* g`isJE^ */
} APUSQUARE;

/* Opg (Ch2) */
typedef struct _APUTRIANGLE {
	APULENGTH length;		/* OXJE^ */
	APULINEAR linear;		/* jAJE^ */
	short wavelength;		/* g ($000..$7ff) */
	unsigned char waveindex;	/* g`CfNX (0..31) */
	int wavecount;			/* g`isJE^ */
} APUTRIANGLE;

/* mCY (Ch3) */
typedef struct _APUNOISE {
	APULENGTH length;		/* OXJE^ */
	APUENVELOPE envelope;		/* Gx[v */
	unsigned char mode;		/* 0 = long mode, 1 = short mode */
	unsigned short lfsr;		/* LFSR (Linear Feedback Shift Register) */
	short wavelength;		/* g ($000..$fff) */
	int wavecount;			/* g`isJE^ */
} APUNOISE;

/* f^PCM (Ch4) */
typedef struct _APUDMC {
	int dummy;
} APUDMC;

typedef struct _APU {
	unsigned char regs[0x20];	/* WX^ ($4000-$401F) */
	//
	APUSQUARE square[2];		/* Ch1-2: `g */
	APUTRIANGLE triangle;		/*   Ch3: Opg */
	APUNOISE noise;			/*   Ch4: mCY */
	APUDMC dmc;			/*   Ch5: f^PCM */
	//
	unsigned char irq;		/*   D7: DMC IRQ   */
					/*   D6: Frame IRQ */
					/* D5-0: 0         */
	unsigned char fc_250_240;	/* 250Hz->240HzϊpJE^ */
	unsigned char framecount;	/* 240HzJE^[4->]3->2->1->0 */
} APU;

extern APU apu;

/****************************************************************************
 *
 ****************************************************************************/

extern const unsigned char apu_square_duty_table[4];
extern const char apu_triangle_waveform_table[32];
extern const short apu_noise_wavelength_table[16];
extern const unsigned char apu_length_count_table[32];

void apu_do_length(APULENGTH* length);
void apu_do_linear(APULINEAR* linear);
void apu_do_envelope(APUENVELOPE* envelope);
void apu_do_sweep(APUSWEEP* sweep);

void apu_square_mix(APUSQUARE* square, short wbuff[/*64*/]);
void apu_triangle_mix(APUTRIANGLE* triangle, short wbuff[/*64*/]);
void apu_noise_mix(APUNOISE* noise, short wbuff[/*64*/]);
void apu_dmc_mix(APUDMC* dmc, short wbuff[/*64*/]);
void apu_volume_shift(short wbuff[/*APUBUFLEN*/]);

/****************************************************************************
 *
 ****************************************************************************/

void apu_reset();
void apu_write(int addr, int data);
int apu_read();
void apu_mix(short wbuff[/*APUBUFLEN*/]);

#endif /*__CLIP_APU_H__*/
