/*	
 *	framapu.c
 *
 *	P/ECE APU (RICOH RP2A03) Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2005 Naoyuki Sawa
 *
 *	* Sun Feb 06 18:39:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *
 ****************************************************************************/

void
apu_square_mix(APUSQUARE* square, short wbuff[/*64*/])
{
	APULENGTH* length = &square->length;
	APUENVELOPE* envelope = &square->envelope;
	APUSWEEP* sweep = &square->sweep;
	//
	int i;
	int volume;
	int wavelength;
	int waveindex;
	int wavecount;
	int duty;
	int output;

	/* OXJE^0Ȃ΁A܂B */
	if(!length->count) {
		return;
	}

	/* {[0Ȃ΁A܂B */
	volume = envelope->volume;
	if(!volume) {
		return;
	}

	/* g8ȉ܂0x7ffȏȂ΁A܂B */
	wavelength = sweep->wavelength;
	if(wavelength <= 8 || wavelength >= 0x7ff) {
		return;
	}
	wavelength = (wavelength + 1) * SPEAKER_FREQUENCY;
	// min(square->wavelength) = 9 łA
	//   (wavelength + 1) * SPEAKER_FREQUENCY = 160000
	//   1.79MHz / 144000 = 11.2
	// 1Tv̍őXV񐔂12ȉłB
	// P/ECEłԂɍƎv̂ŁA~b^݂͐܂B

	duty = square->duty;

	waveindex = square->waveindex; /* o */
	wavecount = square->wavecount; /* o */

	output = (waveindex < duty) ? volume : -volume;

	i = 64;
	do {
		wavecount -= APU_CLOCK;
		if(wavecount < 0) {
			do {
				waveindex++;
			} while((wavecount += wavelength) < 0);
			waveindex &= 15;
			output = (waveindex < duty) ? volume : -volume;
		}
		*wbuff++ += output;
	} while(--i);

	square->waveindex = waveindex; /* ߂ */
	square->wavecount = wavecount; /* ߂ */
}

void
apu_triangle_mix(APUTRIANGLE* triangle, short wbuff[/*64*/])
{
	APULENGTH* length = &triangle->length;
	APULINEAR* linear = &triangle->linear;
	//
	int i;
	int wavelength;
	int waveindex;
	int wavecount;
	int output;

	/* OXJE^0Ȃ΁A܂B */
	if(!length->count) {
		return;
	}

	/* jAJE^0Ȃ΁A܂B */
	if(!linear->count) {
		return;
	}

	/* g8ȉ܂0x7ffȏȂ΁A܂B */
	wavelength = triangle->wavelength;
	if(wavelength <= 8 || wavelength >= 0x7ff) { /* dl */
		return;
	}
	wavelength = (wavelength + 1) * SPEAKER_FREQUENCY;
	// min(square->wavelength) = 9 łA
	//   (wavelength + 1) * SPEAKER_FREQUENCY = 160000
	//   1.79MHz / 144000 = 11.2
	// 1Tv̍őXV񐔂12ȉłB
	// P/ECEłԂɍƎv̂ŁA~b^݂͐܂B

	waveindex = triangle->waveindex; /* o */
	wavecount = triangle->wavecount; /* o */

	output = apu_triangle_waveform_table[waveindex];

	i = 64;
	do {
		wavecount -= APU_CLOCK;
		if(wavecount < 0) {
			do {
				waveindex++;
			} while((wavecount += wavelength) < 0);
			waveindex &= 31;
			output = apu_triangle_waveform_table[waveindex];
		}
		*wbuff++ += output;
	} while(--i);

	triangle->waveindex = waveindex; /* ߂ */
	triangle->wavecount = wavecount; /* ߂ */
}

void
apu_noise_mix(APUNOISE* noise, short wbuff[/*64*/])
{
	APULENGTH* length = &noise->length;
	APUENVELOPE* envelope = &noise->envelope;
	//
	int i;
	int v;
	int volume;
	int wavelength;
	int wavecount;
	int mode;
	int lfsr;
	int output;

	/* OXJE^0Ȃ΁A܂B */
	if(!length->count) {
		return;
	}

	/* {[0Ȃ΁A܂B */
	volume = envelope->volume;
	if(!volume) {
		return;
	}

	wavelength = noise->wavelength;
	wavelength = wavelength * SPEAKER_FREQUENCY;
	// min(noise->wavelength) = 4 łA
	//   wavelength * SPEAKER_FREQUENCY = 160000
	//   1.79MHz / 144000 = 28.0
	// 1Tv̍őXV񐔂29ȉłB
	// ͂Ȃ茵ł...~b^KvH

	mode = noise->mode;

	wavecount = noise->wavecount; /* o */
	lfsr      = noise->lfsr;      /* o */

	output = (lfsr & 1) ? volume : -volume;

	i = 64;
	do {
		wavecount -= APU_CLOCK;
		if(wavecount < 0) {
			do {
				if(!mode) {
					v = ((lfsr>>0) ^ (lfsr>>1)) & 1;
				} else {
					v = ((lfsr>>0) ^ (lfsr>>6)) & 1;
				}
				lfsr = (lfsr >> 1) | (v << 14);
			} while((wavecount += wavelength) < 0);
			output = !(lfsr & 1) ? volume : -volume;
		}
		*wbuff++ += output;
	} while(--i);
	noise->wavecount = wavecount; /* ߂ */
	noise->lfsr      = lfsr;      /* ߂ */
}

void
apu_dmc_mix(APUDMC* dmc, short wbuff[/*64*/])
{
	/* TODO: */
}

/****************************************************************************
 *
 ****************************************************************************/

/* t8bit~5ch = t10.5bit */
#define APU_VOLUME_SHIFT ((16 - 11) + 2/**/)

#ifndef APU_ASM

void
apu_volume_shift(short wbuff[/*APUBUFLEN*/])
{
	int i;
	int v;

	i = APUBUFLEN;
	do {
		v = *wbuff;
		v <<= APU_VOLUME_SHIFT;
		if(v >  32767) v =  32767;
		if(v < -32768) v = -32768;
		*wbuff++ = v;
	} while(--i);
}

#else /*APU_ASM*/

asm("
	.code
	.align 1
	.global apu_volume_shift
apu_volume_shift:
	; %r12 = wbuff
	xld.w %r13, 320			; APUBUFLEN = 320
	xld.w %r4, 32767
	not %r5, %r4			; -32768
apu_volume_shift_DO:
	ld.h %r6, [%r12]
	xsll %r6, 7			; APU_VOLUME_SHIFT = (16-11)+2 = 7
	cmp %r6, %r4
	jrle.d 3
	cmp %r6, %r5			; *delay*
	ld.w %r6, %r4			; (skip?)
	jrge.d 3
	sub %r13, 1			; *delay*
	ld.w %r6, %r5			; (skip?)
	ld.h [%r12]+, %r6
	xjrne apu_volume_shift_DO
	ret
");

#endif /*APU_ASM*/
